﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace QR_Encoder1
{
    class RS_Encode
    {
        private string irreduziblesPolynom;
        private int irreduziblesPolynom_Dez;

        private string nachrichtDaten; 
            /*   Trenner ist Leerzeichen
             z.B.    "120 33    47  210 80   37"   
             oder    "  13  11 22   3   213 88    129  23"  
             oder    "13 33   2  44 251   199  98 "  
            */
        public int[] nachricht, originalNachricht;

        
        private int eccAnzahl;

        public int[] gen;

        private int[] log2Feld = new int[256];
        private int[] pot2Feld = new int[256];
        private int[] invFeld = new int[256];

        private void vorbereitung()
        {
            // Vorbereitung der Hilfsrechenwerte ... besser hier im Konstruktor 
            for (int i = 1; i <= 255; i++)
            {
                int wert = xHyR(2, i);
                log2Feld[wert] = i;
                pot2Feld[i] = wert;
            }
            for (int i = 1; i <= 255; i++)
            {
                invFeld[i] = xHyR(2, 255 - log2Feld[i]);
            }
        }

        public RS_Encode()
        {
            irreduziblesPolynom = "100011101"; // bei QR-Codes normal
            irreduziblesPolynom_Dez = Convert.ToInt16(irreduziblesPolynom, 2);
            // QR-Code: normal 285 (binär 100011101)

            vorbereitung();
        }

       
        public RS_Encode(string irreduziblesPoly)
        {
            irreduziblesPolynom = irreduziblesPoly;
            irreduziblesPolynom_Dez = Convert.ToInt16(irreduziblesPolynom, 2);
             // QR-Code: normal 285 (binär 100011101)

            vorbereitung();
        }

        private int xHoch_y(int a, int b) // veraltet - wird nicht mehr gebraucht
        {
            int wert = a;
            if (b >= 2) for (int i = 2; i <= b; i++) wert = multiplikation(wert, a);
            else { if (b == 0) wert = 1; }
            return wert;
        }


        private int multiplikation(int w1, int w2)
        {   // siehe     h t t p s://en.wikipedia.org/wiki/Finite_field_arithmetic
            int p = 0; /* Produkt der Multiplikation */
            int a = w1;
            int b = w2;
            while ((a != 0) && (b != 0))
            {
                if ((b % 2) > 0) // falls b gerade ist ... Addition von a zum Produkt 
                                 // Endprodukt = Summe aller  a's ... bei geraden b's 
                    p ^= a;      // wir bewegen uns im GF(2^m) --> Addition mit XOR 
                if (a >= 128)    // GF modulo: if a >= 128, then it will overflow when shifted left, so reduce 
                    a = (a << 1) ^ irreduziblesPolynom_Dez;  //  285; 
                                                             // irreversibles Polynom      1 + x ^ 2 + x ^ 3 + x ^ 4 + x ^ 8;     100011101 -- > 285
                                                             // 0x11b; /* XOR with the primitive polynomial x^8 + x^4 + x^3 + x + 1 (0b100011011) --> 
                                                             //            – you can change it but it must be irreducible */
                else
                    a <<= 1; // entspricht  a = a * 2 
                b >>= 1; // entspricht  b = b / 2 
            }

            return p;
        }


        private int xHyR(int x, int y) // besser als xHoch_y
        {
            int erg = 0;
            if (y >= 2)
            {
                int h = xHyR(x, y / 2);
                erg = multiplikation(h, h);
                if (y % 2 != 0)
                {
                    erg = multiplikation(erg, x);
                }
            }
            else
            {
                if (y == 1) erg = x;
                else erg = 1;
            }
            return erg;
        }

        private int[] polynomMultiplikationBinär(int[] a, int[] b, int moduloWert) // nur relevant bei moduloWert>1
        {
            // a[0] müssen die Einer darstellen, a[1] zu x, a[2] zu x^2 etc. 
            // 3x^5 + 12x^3 - 5x^2 +  x - 99   =   99   + x       - 5x^2 + 12x^3 + 3x^5  
            //                          --->  a[0]=99, a[1]=1, ... -5,     12,   0, 3
            // c = a * b;
            int[] c = new int[a.Length + b.Length];
            for (int i = 0; i < c.Length; i++) c[i] = 0;

            for (int i = 0; i < b.Length; i++)
                for (int j = 0; j < a.Length; j++)
                {
                    int wert = 0;
                    if ((a[j] * b[i]) != 0)
                    {
                        int hz1 = log2Feld[a[j]], hz2 = log2Feld[b[i]];
                        int hz = (hz1 + hz2) % 255;
                        wert = xHyR(2, hz);
                    }

                    c[i + j] = c[i + j] ^ wert;  // a[j] * b[i];
                }
            if (moduloWert > 1) for (int i = 0; i < c.Length; i++) c[i] = c[i] % moduloWert;
            return c;
        }



        private int[] erstelleGeneratorpolynom(int k)
        {
            int[] gen;
            int[] a = { 1, 1 };
            int[] b = { 2, 1 };
            int[] c;
            int letztWert = 0;

            if (k > 1)
            {
                c = polynomMultiplikationBinär(a, b, 256); 
                letztWert = 0;
                for (int j = c.Length - 1; j >= 0; j--) if (c[j] > 0) { letztWert = j; j = -1; }
                for (int j = 0; j <= letztWert; j++)
                {
                    if (j >= a.Length) { Array.Resize(ref a, a.Length + 1); }
                    a[j] = c[j];
                }
            }

            if (k > 2)
            {
                for (int i = 1; i <= k - 2; i++)
                {
                    b[0] = multiplikation(b[0], 2);
                    b[1] = 1;
                    c = polynomMultiplikationBinär(a, b, 256); 
                    letztWert = 0;
                    for (int j = c.Length - 1; j >= 0; j--) if (c[j] > 0) { letztWert = j; j = -1; }
                    for (int j = 0; j <= letztWert; j++)
                    {
                        if (j >= a.Length) Array.Resize(ref a, a.Length + 1);
                        a[j] = c[j];
                    }

                }
            }
            letztWert = 0;
            for (int i = a.Length - 1; i >= 0; i--) if (a[i] > 0) { letztWert = i; i = -1; }
            Array.Resize(ref a, letztWert + 1);
            gen = a;

            return gen;
        }

        private int[] polynomDivisionRest(int[] aOrg, int[] b)
        {
            int[] a = new int[aOrg.Length];
            for (int i = 0; i < a.Length; i++) a[i] = aOrg[i];

            int hoechstBit = 0;
            int[] h = new int[a.Length];

            for (int i = a.Length - 1; i >= 0; i--) if (a[i] > 0) { hoechstBit = i; i = -1; }

            while (hoechstBit >= b.Length - 1)
            {
                int diff = hoechstBit - (b.Length - 1);
                for (int i = b.Length - 1; i >= 0; i--) h[i + diff] = multiplikation(a[hoechstBit], b[i]);
                for (int i = diff - 1; i >= 0; i--) h[i] = 0; // rechts mit Nullen auffüllen

                for (int i = hoechstBit; i >= 0; i--) a[i] = a[i] ^ h[i]; // XOR wegen "modulo 2" -Rechnung;   ZwischenRest in a
                for (int i = a.Length - 1; i >= 0; i--) if (a[i] > 0) { hoechstBit = i; i = -1; }
            }
            return a;
        }

        public int[] RSEncode(string datenBytes, int anzahlECC, string irreduziblesPoly)
        {
            irreduziblesPolynom = irreduziblesPoly;
            irreduziblesPolynom_Dez = Convert.ToInt16(irreduziblesPolynom, 2);  
                                   // QR-Code: normal 285 (binär 100011101)

            // Nachrichten in int-Array umwandeln:
            nachrichtDaten = datenBytes;
            eccAnzahl = anzahlECC;

            string s = nachrichtDaten.Trim();

            string[] datenStrings = s.Split(' ');
            int[] a = new int[datenStrings.Length];
            int zw = 0, h1=-1;
            for (int i=0; i< datenStrings.Length; i++)
            {
                if (datenStrings[i].Trim() != "")
                {
                    try { h1 = Convert.ToInt16(datenStrings[i].Trim()); } catch { h1=-1; }
                    if (h1 >= 0) { a[zw] = h1; zw++; }
                }
            }
            Array.Resize(ref a, zw);
            nachricht = a;
            
            // Generatorpolynom erstellen:
            int k = eccAnzahl;
            gen = erstelleGeneratorpolynom(k);
            
            // Nachrichten erweitern auf GerneratorpolynomLänge und Reihenfolge umdrehen
            int origLaenge = nachricht.Length;
            Array.Resize(ref nachricht, origLaenge + k);
            for (int i = origLaenge + k - 1; i >= origLaenge; i--) nachricht[i] = 0;

            int[] nachrichtHilf = new int[nachricht.Length];
            for (int i = 0; i < nachricht.Length; i++) nachrichtHilf[nachricht.Length - 1 - i] = nachricht[i];
            for (int i = 0; i < nachricht.Length; i++) nachricht[i] = nachrichtHilf[i];

            originalNachricht = new int[nachricht.Length];
            for (int i = 0; i < nachricht.Length; i++) originalNachricht[i] = nachricht[i];

            // Polynomdivision
            int[] rest;
            
            rest = polynomDivisionRest(nachricht, gen);

            return rest; 
        }

    }
}
